# Generated by roxygen2: do not edit by hand

S3method(plot,layout)
S3method(print,grip_gmds_layout)
S3method(print,grip_misf)
export(build.misf)
export(build.weighted.misf)
export(compare.layouts)
export(cube.asymmetric.cavities.surface.embedding)
export(cube.asymmetric.cavities.surface.graph)
export(cube.channel.network.surface.embedding)
export(cube.channel.network.surface.graph)
export(cube.periodic.tunnels.surface.embedding)
export(cube.periodic.tunnels.surface.graph)
export(cylinder.surface.embedding)
export(cylinder.surface.graph)
export(edge.kk)
export(edge.length.density.stiffness)
export(edge.repulsive.stage)
export(edge.repulsive.state)
export(edges.cube)
export(edges.cube.asymmetric.cavities)
export(edges.cube.channel.network)
export(edges.cube.periodic.tunnels)
export(edges.cycle)
export(edges.cylinder)
export(edges.irregular.annulus)
export(edges.irregular.ball)
export(edges.irregular.double.torus)
export(edges.irregular.pair.of.pants)
export(edges.irregular.shell)
export(edges.irregular.sphere)
export(edges.irregular.torus)
export(edges.kary.tree)
export(edges.menger.sponge)
export(edges.mesh)
export(edges.occupied.mesh)
export(edges.path)
export(edges.recursive.cube.mask)
export(edges.recursive.mask.grid)
export(edges.recursive.tetrahedron.mask)
export(edges.recursive.triangle.mask)
export(edges.sierpinski.carpet)
export(edges.sierpinski.tetrahedron)
export(edges.sierpinski.triangle)
export(edges.sphere)
export(edges.torus)
export(edges.triangulated.annulus)
export(edges.triangulated.pair.of.pants)
export(edges.triangulated.polyhedron)
export(edges.vicsek)
export(geodesic.kk)
export(geometry.diagnostics)
export(globalrep.grip)
export(globalrep.weighted.grip)
export(gmds.result)
export(graph.riemannian.star.structure)
export(grip)
export(grip.build.misf)
export(grip.build.misf.weighted)
export(grip.compare.layouts)
export(grip.edge.length.density.stiffness)
export(grip.edge.repulsive.state)
export(grip.geometry.diagnostics)
export(grip.gmds.layout.result)
export(grip.layout)
export(grip.layout.globalrep)
export(grip.layout.globalrep.weighted)
export(grip.layout.legacy)
export(grip.layout.trace)
export(grip.layout.trace.legacy)
export(grip.layout.trace.weighted)
export(grip.layout.weighted)
export(grip.layout.weighted.nd)
export(grip.metric.mds.layout)
export(grip.optimize.edge.gkk.layout)
export(grip.optimize.edge.isometric.layout)
export(grip.optimize.edge.kk.layout)
export(grip.optimize.edge.repulsive.stage)
export(grip.optimize.geodesic.kk)
export(grip.optimize.kernel.gram.gkk.layout)
export(grip.optimize.landmark.geodesic.kk)
export(grip.optimize.misf.geodesic.kk)
export(grip.optimize.repulsive.stage)
export(grip.params.from.summary)
export(grip.plot)
export(grip.prepare.edge.kk)
export(grip.prepare.geodesic.kk)
export(grip.prepare.graph.geodesic.mds)
export(grip.prepare.landmark.geodesic.kk)
export(grip.prepare.misf.geodesic.kk)
export(grip.project.3d)
export(grip.repulsive.state)
export(grip.score.geodesic.kk)
export(grip.score.gmds.layout)
export(grip.score.landmark.geodesic.kk)
export(grip.score.layout)
export(grip.score.misf.geodesic.kk)
export(gripui_app)
export(gripui_family_app)
export(gripui_graph_family_catalog)
export(gripui_project)
export(gripui_project_from_compare)
export(gripui_project_from_dir)
export(gripui_validate_project)
export(irregular.annulus.surface.embedding)
export(irregular.annulus.surface.graph)
export(irregular.ball.solid.embedding)
export(irregular.ball.solid.graph)
export(irregular.double.torus.surface.embedding)
export(irregular.double.torus.surface.graph)
export(irregular.pair.of.pants.surface.embedding)
export(irregular.pair.of.pants.surface.graph)
export(irregular.rectangle.param.coords)
export(irregular.rectangle.surface.embedding)
export(irregular.rectangle.surface.graph)
export(irregular.shell.solid.embedding)
export(irregular.shell.solid.graph)
export(irregular.sphere.surface.embedding)
export(irregular.sphere.surface.graph)
export(irregular.torus.surface.embedding)
export(irregular.torus.surface.graph)
export(kary.tree.weighted.graph)
export(keep.asymmetric.notches)
export(keep.periodic.holes)
export(keep.slit.channels)
export(keep.staggered.windows)
export(kernel.gram.gkk)
export(landmark.geodesic.kk)
export(legacy.grip)
export(mask.asymmetric.holes)
export(mask.border)
export(mask.corner)
export(mask.cross)
export(mask.cube.asymmetric.cavities)
export(mask.cube.channel.network)
export(mask.cube.periodic.tunnels)
export(mask.tetrahedron.classic)
export(mask.tetrahedron.corner.missing)
export(mask.triangle.bridge)
export(mask.triangle.classic)
export(menger.sponge.surface.embedding)
export(menger.sponge.surface.graph)
export(mesh.surface.embedding)
export(mesh.surface.graph)
export(metric.mds)
export(misf.geodesic.kk)
export(occupied.mesh.surface.embedding)
export(occupied.mesh.surface.graph)
export(params.from.summary)
export(plot.layout)
export(prepare.edge.kk)
export(prepare.geodesic.kk)
export(prepare.graph.geodesic.mds)
export(prepare.landmark.geodesic.kk)
export(prepare.misf.geodesic.kk)
export(project.3d)
export(recursive.cube.mask.surface.embedding)
export(recursive.cube.mask.surface.graph)
export(recursive.mask.grid.surface.embedding)
export(recursive.mask.grid.surface.graph)
export(recursive.tetrahedron.mask.surface.embedding)
export(recursive.tetrahedron.mask.surface.graph)
export(recursive.triangle.mask.surface.embedding)
export(recursive.triangle.mask.surface.graph)
export(repulsive.stage)
export(repulsive.state)
export(run_gripui)
export(run_gripui_family)
export(sampled.rectangle.param.coords)
export(sampled.rectangle.surface.embedding)
export(sampled.rectangle.surface.graph)
export(sampled.rectangle.surface.graphs)
export(score.geodesic.kk)
export(score.gmds)
export(score.landmark.geodesic.kk)
export(score.layout)
export(score.misf.geodesic.kk)
export(sierpinski.carpet.surface.embedding)
export(sierpinski.carpet.surface.graph)
export(sierpinski.tetrahedron.surface.embedding)
export(sierpinski.tetrahedron.surface.graph)
export(sierpinski.triangle.surface.embedding)
export(sierpinski.triangle.surface.graph)
export(sphere.surface.embedding)
export(sphere.surface.graph)
export(torus.surface.embedding)
export(torus.surface.graph)
export(trace.grip)
export(trace.legacy.grip)
export(trace.weighted.grip)
export(triangulated.annulus.surface.embedding)
export(triangulated.annulus.surface.graph)
export(triangulated.pair.of.pants.surface.embedding)
export(triangulated.pair.of.pants.surface.graph)
export(triangulated.polyhedron.surface.embedding)
export(triangulated.polyhedron.surface.graph)
export(vicsek.surface.embedding)
export(vicsek.surface.graph)
export(weighted.grip)
export(weighted.grip.nd)
importFrom(Rcpp,evalCpp)
importFrom(graphics,points)
importFrom(graphics,segments)
useDynLib(grip, .registration = TRUE)
