% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gripui-project.R
\name{gripui_project_from_compare}
\alias{gripui_project_from_compare}
\title{Convert `compare.layouts()` output into a `gripui_project`}
\usage{
gripui_project_from_compare(
  compare_obj,
  graph = NULL,
  vertex_data = NULL,
  graph_info = NULL,
  title = NULL
)
}
\arguments{
\item{compare_obj}{Result of `compare.layouts()`.}

\item{graph}{Optional graph object, named graph list, or path to a graph RDS.
Defaults to `NULL` when only catalog exploration is needed.}

\item{vertex_data}{Optional vertex metadata added to `graph`.}

\item{graph_info}{Optional graph-level metadata added to `graph`.}

\item{title}{Optional project title.}
}
\value{
A `gripui_project`.
}
\description{
Convert `compare.layouts()` output into a `gripui_project`
}
\examples{
edges <- edges.path(5)
cmp <- compare.layouts(
  edges = edges,
  n = 5,
  candidates = "default",
  seeds = 1L,
  return.layouts = TRUE
)
graph <- list(
  adj_list = list(2L, c(1L, 3L), c(2L, 4L), c(3L, 5L), 4L)
)
project <- gripui_project_from_compare(cmp, graph = graph, title = "Path compare")
nrow(project$layouts)
}
