% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grip_quality.R
\name{grip.score.layout}
\alias{grip.score.layout}
\title{Deprecated layout scoring name}
\usage{
grip.score.layout(...)
}
\arguments{
\item{...}{Arguments passed to [score.layout()].}
}
\value{
See [score.layout()].
}
\description{
This long name is deprecated. Use [score.layout()] instead.
}
