% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grip_quality.R
\name{grip.compare.layouts}
\alias{grip.compare.layouts}
\title{Deprecated layout comparison name}
\usage{
grip.compare.layouts(...)
}
\arguments{
\item{...}{Arguments passed to [compare.layouts()].}
}
\value{
See [compare.layouts()].
}
\description{
This long name is deprecated. Use [compare.layouts()] instead.
}
