% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gripui-app.R
\name{run_gripui}
\alias{run_gripui}
\title{Run the `gripui` Shiny application}
\usage{
run_gripui(
  project,
  host = "127.0.0.1",
  port = getOption("shiny.port"),
  launch.browser = interactive(),
  auto.stop.after = NULL,
  ...
)
}
\arguments{
\item{project}{A `gripui_project`.}

\item{host}{Host passed to `shiny::runApp()`.}

\item{port}{Port passed to `shiny::runApp()`.}

\item{launch.browser}{Whether to launch a browser.}

\item{auto.stop.after}{Optional delay, in seconds, after which the app stops
itself. This is mainly useful for automated examples and tests.}

\item{...}{Additional arguments passed to `shiny::runApp()`.}
}
\value{
Invisibly returns the result of `shiny::runApp()`.
}
\description{
Run the `gripui` Shiny application
}
\examples{
\dontshow{if (local({ old <- getOption("rgl.useNULL"); options(rgl.useNULL = TRUE); on.exit(options(rgl.useNULL = old), add = TRUE); requireNamespace("shiny", quietly = TRUE) && requireNamespace("bslib", quietly = TRUE) && requireNamespace("DT", quietly = TRUE) && requireNamespace("htmltools", quietly = TRUE) && requireNamespace("rgl", quietly = TRUE) && requireNamespace("later", quietly = TRUE) })) withAutoprint(\{ # examplesIf}
graph <- list(adj_list = list(2L, c(1L, 3L), 2L))
layouts <- data.frame(
  candidate = "toy.layout",
  stage = "layout",
  seed = 1L,
  status = "ok",
  stringsAsFactors = FALSE
)
project <- gripui_project(graph = graph, layouts = layouts, title = "Toy project")
run_gripui(
  project,
  launch.browser = FALSE,
  quiet = TRUE,
  auto.stop.after = 0.1
)
\dontshow{\}) # examplesIf}
}
