% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gripui-family-app.R
\name{gripui_family_app}
\alias{gripui_family_app}
\title{Build the graph-family geometry explorer Shiny application}
\usage{
gripui_family_app(
  catalog = gripui_graph_family_catalog(),
  title = "Graph Family Geometry Explorer",
  subtitle = "Interactive geometry browser for synthetic benchmark families."
)
}
\arguments{
\item{catalog}{Family catalog, usually `gripui_graph_family_catalog()`.}

\item{title}{Application title.}

\item{subtitle}{Optional subtitle shown in the sidebar.}
}
\value{
A `shiny.appobj`.
}
\description{
Build the graph-family geometry explorer Shiny application
}
\examples{
\dontshow{if (local({ old <- getOption("rgl.useNULL"); options(rgl.useNULL = TRUE); on.exit(options(rgl.useNULL = old), add = TRUE); requireNamespace("shiny", quietly = TRUE) && requireNamespace("bslib", quietly = TRUE) && requireNamespace("rgl", quietly = TRUE) })) withAutoprint(\{ # examplesIf}
app <- gripui_family_app()
inherits(app, "shiny.appobj")
\dontshow{\}) # examplesIf}
}
