% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gripui-project.R
\name{gripui_validate_project}
\alias{gripui_validate_project}
\title{Validate a `gripui_project`}
\usage{
gripui_validate_project(project)
}
\arguments{
\item{project}{Object to validate.}
}
\value{
Invisibly returns `TRUE` when validation succeeds.
}
\description{
Validate a `gripui_project`
}
\examples{
layouts <- data.frame(
  candidate = "test.layout",
  stage = "layout",
  seed = 1L,
  status = "ok",
  stringsAsFactors = FALSE
)
project <- gripui_project(graph = NULL, layouts = layouts, title = "Validation example")
gripui_validate_project(project)
}
