% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gripui-project.R
\name{gripui_project}
\alias{gripui_project}
\title{Create a normalized `gripui` project object}
\usage{
gripui_project(
  graph = NULL,
  layouts,
  title = NULL,
  subtitle = NULL,
  notes = NULL
)
}
\arguments{
\item{graph}{Optional graph object used by layout viewers. This may be a
single list containing fields such as `adj_list`, `weight_list`,
`vertex_data`, and `graph_info`, a named list of such graph objects for
multi-stage projects, or `NULL` when only catalog exploration is needed.}

\item{layouts}{Data frame with one row per realized layout.}

\item{title}{Optional project title.}

\item{subtitle}{Optional subtitle.}

\item{notes}{Optional notes string.}
}
\value{
An object of class `gripui_project`.
}
\description{
Create a normalized `gripui` project object
}
\examples{
graph <- list(adj_list = list(2L, c(1L, 3L), c(2L, 4L), 3L))
layouts <- data.frame(
  candidate = "path.default",
  stage = "layout",
  seed = 1L,
  status = "ok",
  stringsAsFactors = FALSE
)
project <- gripui_project(graph = graph, layouts = layouts, title = "Path graph")
project$meta$title
}
