% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grip_quality.R
\name{grip.prepare.edge.kk}
\alias{grip.prepare.edge.kk}
\title{Deprecated edge-KK preparation name}
\usage{
grip.prepare.edge.kk(...)
}
\arguments{
\item{...}{Arguments passed to [prepare.edge.kk()].}
}
\value{
See [prepare.edge.kk()].
}
\description{
This long name is deprecated. Use [prepare.edge.kk()] instead.
}
