% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gripui-app.R
\name{gripui_app}
\alias{gripui_app}
\title{Build the `gripui` Shiny application}
\usage{
gripui_app(project)
}
\arguments{
\item{project}{A `gripui_project`.}
}
\value{
A `shiny.appobj`.
}
\description{
Build the `gripui` Shiny application
}
\examples{
\dontshow{if (local({ old <- getOption("rgl.useNULL"); options(rgl.useNULL = TRUE); on.exit(options(rgl.useNULL = old), add = TRUE); requireNamespace("shiny", quietly = TRUE) && requireNamespace("bslib", quietly = TRUE) && requireNamespace("DT", quietly = TRUE) && requireNamespace("htmltools", quietly = TRUE) && requireNamespace("rgl", quietly = TRUE) })) withAutoprint(\{ # examplesIf}
graph <- list(adj_list = list(2L, c(1L, 3L), 2L))
layouts <- data.frame(
  candidate = "toy.layout",
  stage = "layout",
  seed = 1L,
  status = "ok",
  stringsAsFactors = FALSE
)
project <- gripui_project(graph = graph, layouts = layouts, title = "Toy project")
app <- gripui_app(project)
inherits(app, "shiny.appobj")
\dontshow{\}) # examplesIf}
}
