% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deprecated_api.R
\name{deprecated-grip-api}
\alias{deprecated-grip-api}
\alias{grip.layout}
\alias{grip.layout.globalrep}
\alias{grip.layout.globalrep.weighted}
\alias{grip.layout.legacy}
\alias{grip.layout.trace}
\alias{grip.layout.trace.legacy}
\alias{grip.layout.trace.weighted}
\alias{grip.layout.weighted}
\alias{grip.layout.weighted.nd}
\alias{grip.plot}
\alias{grip.project.3d}
\alias{grip.metric.mds.layout}
\alias{grip.optimize.kernel.gram.gkk.layout}
\alias{grip.score.gmds.layout}
\alias{grip.gmds.layout.result}
\alias{grip.edge.length.density.stiffness}
\alias{grip.prepare.landmark.geodesic.kk}
\alias{grip.prepare.geodesic.kk}
\alias{grip.prepare.graph.geodesic.mds}
\alias{grip.prepare.misf.geodesic.kk}
\alias{grip.score.landmark.geodesic.kk}
\alias{grip.score.geodesic.kk}
\alias{grip.score.misf.geodesic.kk}
\alias{grip.optimize.landmark.geodesic.kk}
\alias{grip.optimize.geodesic.kk}
\alias{grip.optimize.misf.geodesic.kk}
\alias{grip.optimize.edge.repulsive.stage}
\alias{grip.optimize.repulsive.stage}
\alias{grip.edge.repulsive.state}
\alias{grip.repulsive.state}
\alias{grip.build.misf.weighted}
\alias{grip.build.misf}
\alias{grip.geometry.diagnostics}
\alias{grip.params.from.summary}
\title{Deprecated long-form API names}
\usage{
grip.layout(...)

grip.layout.globalrep(...)

grip.layout.globalrep.weighted(...)

grip.layout.legacy(...)

grip.layout.trace(...)

grip.layout.trace.legacy(...)

grip.layout.trace.weighted(...)

grip.layout.weighted(...)

grip.layout.weighted.nd(...)

grip.plot(...)

grip.project.3d(...)

grip.metric.mds.layout(...)

grip.optimize.kernel.gram.gkk.layout(...)

grip.score.gmds.layout(...)

grip.gmds.layout.result(...)

grip.edge.length.density.stiffness(...)

grip.prepare.landmark.geodesic.kk(...)

grip.prepare.geodesic.kk(...)

grip.prepare.graph.geodesic.mds(...)

grip.prepare.misf.geodesic.kk(...)

grip.score.landmark.geodesic.kk(...)

grip.score.geodesic.kk(...)

grip.score.misf.geodesic.kk(...)

grip.optimize.landmark.geodesic.kk(...)

grip.optimize.geodesic.kk(...)

grip.optimize.misf.geodesic.kk(...)

grip.optimize.edge.repulsive.stage(...)

grip.optimize.repulsive.stage(...)

grip.edge.repulsive.state(...)

grip.repulsive.state(...)

grip.build.misf.weighted(...)

grip.build.misf(...)

grip.geometry.diagnostics(...)

grip.params.from.summary(...)
}
\arguments{
\item{...}{Arguments passed to the replacement function.}
}
\value{
The replacement function's return value.
}
\description{
These long names are deprecated in favor of the shorter canonical API.
}
