% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/graph_helpers.R
\name{tetrahedron_mask_helpers}
\alias{tetrahedron_mask_helpers}
\alias{mask.tetrahedron.classic}
\alias{mask.tetrahedron.corner.missing}
\title{Tetrahedron mask helpers for recursive gasket families}
\usage{
mask.tetrahedron.classic()

mask.tetrahedron.corner.missing(
  omit = c("apex", "base_left", "base_right", "base_back")
)
}
\arguments{
\item{omit}{Which tetrahedron corner is omitted in
\code{mask.tetrahedron.corner.missing()}.}
}
\value{
A named logical vector of length 4.
}
\description{
Convenience constructors for the four-corner tetrahedron masks used by
\code{\link{edges.recursive.tetrahedron.mask}()} and
\code{\link{recursive.tetrahedron.mask.surface.graph}()}. The slots
correspond to the four corner subtetrahedra created by one barycentric
refinement of a tetrahedron.
}
\details{
The returned mask entries are ordered and named as \code{base_left},
\code{base_right}, \code{base_back}, and \code{apex}.
}
