% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gripui-family-app.R
\name{run_gripui_family}
\alias{run_gripui_family}
\title{Run the graph-family geometry explorer Shiny application}
\usage{
run_gripui_family(
  catalog = gripui_graph_family_catalog(),
  title = "Graph Family Geometry Explorer",
  subtitle = "Interactive geometry browser for synthetic benchmark families.",
  host = "127.0.0.1",
  port = getOption("shiny.port"),
  launch.browser = interactive(),
  auto.stop.after = NULL,
  ...
)
}
\arguments{
\item{catalog}{Family catalog, usually `gripui_graph_family_catalog()`.}

\item{title}{Application title.}

\item{subtitle}{Optional subtitle shown in the sidebar.}

\item{host}{Host passed to `shiny::runApp()`.}

\item{port}{Port passed to `shiny::runApp()`.}

\item{launch.browser}{Whether to launch a browser.}

\item{auto.stop.after}{Optional delay, in seconds, after which the app stops
itself. This is mainly useful for automated examples and tests.}

\item{...}{Additional arguments passed to `shiny::runApp()`.}
}
\value{
Invisibly returns the result of `shiny::runApp()`.
}
\description{
Run the graph-family geometry explorer Shiny application
}
\examples{
\dontshow{if (local({ old <- getOption("rgl.useNULL"); options(rgl.useNULL = TRUE); on.exit(options(rgl.useNULL = old), add = TRUE); requireNamespace("shiny", quietly = TRUE) && requireNamespace("bslib", quietly = TRUE) && requireNamespace("rgl", quietly = TRUE) && requireNamespace("later", quietly = TRUE) })) withAutoprint(\{ # examplesIf}
run_gripui_family(launch.browser = FALSE, quiet = TRUE, auto.stop.after = 0.1)
\dontshow{\}) # examplesIf}
}
