% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gmds_layout_interface.R
\name{grip.optimize.edge.kk.layout}
\alias{grip.optimize.edge.kk.layout}
\alias{grip.optimize.edge.gkk.layout}
\alias{grip.optimize.edge.isometric.layout}
\title{Deprecated edge-KK layout names}
\usage{
grip.optimize.edge.kk.layout(...)

grip.optimize.edge.gkk.layout(...)

grip.optimize.edge.isometric.layout(...)
}
\arguments{
\item{...}{Arguments passed to [edge.kk()].}
}
\value{
A `"grip_gmds_layout"` object.
}
\description{
These long names are deprecated. Use [edge.kk()] instead.
}
